/*
 * linux/include/asm-arm/arch-l7200/uncompress.h
 *
 * Copyright (C) 2000 Steve Hill (sjhill@cotw.com)
 */

static __inline__ void putc(char c)
{
}

static void puts(const char *s)
{
}

static __inline__ void arch_decomp_setup(void)
{
}

#define arch_decomp_wdog()
